unit ConnectionPool;

// TkbmCONNPOOL v. 0.98 (15. Jun 1999)
// =========================================================================
// A database connection pooling package.
//
// BEFORE USING THIS COMPONENT, YOU ARE REQUESTED TO READ ALL OF THIS TEXT
// DOCUMENT AND SPECIALLY TO NOTICE AND ACCEPT THE DISCLAIMER.
//
// Can be used as a demonstration of how to create descendents of TDataSet,
// or as in my case, to allow a program to generate temporary data that can
// be used directly by all data aware controls.
//
// Copyright 1999 Kim Bo Madsen/Optical Services - Scandinavia
// All rights reserved.
//
// You are allowed to used this component in any project for free.
// You are NOT allowed to claim that you have created this component or to
// copy its code into your own component and claim that is was your idea.
//
// Im offering this for free for your convinience, and the ONLY thing I DEMAND
// is to get an E-MAIL about what project this component (or dirived versions)
// is used for. That will be my reward of offering this component for free to you!
//
// You dont need to state my name in your software, although it would be
// appreciated if you do.
//
// If you find bugs or alter the component (f.ex. see suggested enhancements
// further down), please DONT just send the corrected/new code out on the internet,
// but instead send it to me, so I can put it into the official version. You will
// be acredited if you do so.
//
// The components are heavily based on TkbmMemTable and thus that component MUST
// exist to be able to install the TkbmConnPool package.
//
//
// DISCLAIMER
// By using this component or parts theiroff you are accepting the full
// responsibility of the use. You are understanding that the author cant be
// made responsible in any way for any problems occuring using this component.
// You also recognize the author as the creator of this component and agrees
// not to claim otherwize!
//
// Please forward corrected versions (source code ONLY!), comments,
// and emails saying you are using it for this or that project to:
//            kbm@optical.dk
//
// Suggestions for future enhancements:
//
//    A Params editor for Delphi 3 and C++Builder 3.
//
// History:
//
// 0.90: The first Beta release.
//                                                              (20. Apr. 1999)
//
// 0.91: Changed TkbmPooledSession and TkbmPooledDataSet to handle
//       multithreaded sessions correctly.
//                                                              (23. Apr. 1999)
//
// 0.92: Remove freeing of dynamically created TSession.
//       The system maintained SessionList will automatically free it.
//                                                              (29. Apr. 1999)
//
// 0.93: Changed handling of sessionlist due to AV on finalization.
//                                                              (7. May 1999)
//
// 0.94: Changed copydata procedure.
//       Added event for obtaining username and password for database logon.
//
// 0.95: Moved onafteropen functionality to active property handler.
//                                                              (25. May 1999)
//
// 0.96: Created caching mechanisms for TkbmPooledDataset and TkbmConnectionPool.
//
// 0.97: Added time not used on connections. Can be used to eliminate non used connections.
//
// 0.98: Moved active functionality back to onafteropen procedure.
//       Created garbagecollection thread instead of the counter based system.
//       Unwinded some threadlocking regarding caching that could end up in
//       a deadlock.
//=============================================================================

{$ifndef VER100} // CBuilder only
{$ObjExportAll On}
{$ASSERTIONS ON}
{$endif}
//{$G+}

interface

uses
  Windows, Messages, SysUtils, Classes, db,dbtables,kbmMemTable,DsgnIntf,Dialogs;

const
  SEMAPHORE_TIMEOUT=20*60*60*1000;      // 20 minutes should be more than enough.

type
  TkbmConnection = class;
  TkbmConnectionPool = class;
  TkbmPooledSessionList = class;
  TkbmPooledSession = class;
  TkbmPooledQueryDataLink = class;
  TkbmPooledDataSet = class;
  TkbmPooledQuery = class;
  TkbmPooledStoredProc = class;

  TkbmPooledGarbageThread = class(TThread)
  protected
     procedure   DoGarbageCollection;
     procedure   Execute; override;
  public
     ConnectionPool : TkbmConnectionPool;
  end;

  // Cache.
  TkbmCacheItem = class
  private
     FCacheID:longint;
     FTable:TkbmMemTable;
     FParams:TParams;
     FLastAccess:TDateTime;
     FCreated:TDateTime;
  public
     constructor Create(CacheID:longint; DataSet:TDataSet; Records:boolean);
     destructor Destroy; override;
     property LastAccess:TDateTime read FLastAccess write FLastAccess;
     property Created:TDateTime read FCreated write FCreated;
  end;

  TkbmCache = class(TThreadList)
  private
     FOwner:TkbmConnectionPool;
     function GetCount:integer;
  protected
     procedure InternalCleanCache(lst:TList; MaxAge,MaxEntries:integer);
     function InternalLocateItem(lst:TList; CacheID:longint):integer;
  public
     constructor Create(AOwner:TkbmConnectionPool);
     destructor Destroy; override;
     function CalcCRC32(s:string):longint;
     function AddToCache(CacheID:longint; ds:TDataSet; Records:boolean):TkbmCacheItem;
     function GetCached(CacheID:longint; ds:TkbmMemTable):boolean;
     procedure CleanCache(MaxAge,MaxEntries:integer);
     procedure ClearCache;
     property Count:integer read GetCount;
  end;

  // Single connection.
  TkbmConnection = class
  private
     FDatabase:TDatabase;
     FSession:TSession;
     FLockCount:integer;
     FLockedWhen:TDateTime;
     FQueueLength:integer;
     FLastUsed:TDateTime;
     FOwnerID:Longint;

     FSemaphore:THandle;
//     QueueLock:TRTLCriticalSection;
     function GetLocked:boolean;
  public
     constructor Create(ADatabaseTemplate:TDatabase; ConnID:integer; dbParams:TStringList);
     destructor Destroy; override;
     function LockConnection(OwnerID:Longint):TDatabase;
     procedure UnlockConnection;
     property Locked:boolean read GetLocked;
     property LockedWhen:TDateTime read FLockedWhen;
     property QueueLength:integer read FQueueLength;
     property LastUsed:TDateTime read FLastUsed;
     property OwnerID:Longint read FOwnerID;
  end;

  TkbmConnectionGetLogonParams = procedure(DBTemplate:TDatabase; ConnID:Integer; var dbParams:TStringList) of Object;
  TkbmConnectionPoolOnGarbageCollection = procedure(ConnectionPool:TkbmConnectionPool) of object;

  TkbmConnectionPool = class(TComponent)
  private
    { Private declarations }
    FMaxConnections:integer;
    FDBTemplate:TDatabase;
    FOnGetLogonParams:TkbmConnectionGetLogonParams;
    FCache:TkbmCache;
    FMaxCacheAge:integer;
    FMaxCacheEntries:integer;
    FMaxCacheRecordCount:integer;
    FEnableCache:boolean;
    FCacheHits:longint;
    FCacheMisses:longint;
    FConnectionTimeout:integer;
    FUniqueConnectionID:integer;

    FGarbageCollection:boolean;
    FGarbageInterval:integer;
    FGarbageThread : TkbmPooledGarbageThread;   // Garbage collection thread.
    FStopEvent : THandle;       // Used to signal a stop to the garbagecollection thread.
    FOnGarbageCollection:TkbmConnectionPoolOnGarbageCollection;

    DBList:TThreadList;
    function GetCurConnections:integer;
    function GetCurRequests:integer;
    function GetCurConnectionOwnerIDs:string;
    procedure SetGarbageInterval(n:integer);
  protected
  public
    { Public declarations }
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    function GetBestConnection:TkbmConnection;
    procedure TimeoutConnections;
    procedure KillConnections;
    property Cache:TkbmCache read FCache;
    property CacheHits:longint read FCacheHits;
    property CacheMisses:longint read FCacheMisses;
  published
    { Published declarations }
    property MaxConnections:integer read FMaxConnections write FMaxConnections;
    property CurConnections:integer read GetCurConnections;
    property CurConnectionOwnerIDs:string read GetCurConnectionOwnerIDs;
    property CurRequests:integer read GetCurRequests;
    property Database:TDatabase read FDBTemplate write FDBTemplate;
    property OnGetLogonParams:TkbmConnectionGetLogonParams read FOnGetLogonParams write FOnGetLogonParams;
    property MaxCacheAge:integer read FMaxCacheAge write FMaxCacheAge;
    property MaxCacheEntries:integer read FMaxCacheEntries write FMaxCacheEntries;
    property MaxCacheRecordCount:integer read FMaxCacheRecordCount write FMaxCacheRecordCount;
    property EnableCache:boolean read FEnableCache write FEnableCache;
    property ConnectionTimeout:integer read FConnectionTimeout write FConnectionTimeout;
    property GarbageInterval:integer read FGarbageInterval write SetGarbageInterval;
    property GarbageCollection:boolean read FGarbageCollection write FGarbageCollection;
    property OnGarbageCollection:TkbmConnectionPoolOnGarbageCollection read FOnGarbageCollection write FOnGarbageCollection;
  end;

  TkbmPooledSessionList = class(TObject)
  private
     FSessions:TThreadList;
     FUnique:integer;
  public
     constructor Create;
     destructor Destroy; override;

     procedure CloseAll;
     procedure AddSession(Session:TkbmPooledSession);
     procedure DelSession(Session:TkbmPooledSession);
     function SessionByName(const SessionName:string):TkbmPooledSession;

     property Unique:integer read FUnique;
  end;

  TkbmPooledSession = class(TComponent)
  private
     FAutoSessionName:boolean;
     FSessionName:string;
     FThreadSessionName:string;
     FSessionNumber:integer;
     FConnectionPool:TkbmConnectionPool;
     FCon:TkbmConnection;
     FSemaphore:THandle;
     FLockCount:integer;
     FOwnerID:Longint;
     procedure SetSessionName(SessionName:string);
     function GetLocked:boolean;
  protected
     procedure Notification(AComponent: TComponent; Operation: TOperation); override;
     procedure SetSessionNames;
     procedure Loaded; override;
     procedure UpdateThreadSessionName;
     procedure SetName(const NewName: TComponentName); override;
  public
     constructor Create(AOwner:TComponent); override;
     destructor Destroy; override;
     function LockSession(Sync:boolean):TkbmConnection;
     procedure UnlockSession;
     property SessionNumber:integer read FSessionNumber write FSessionNumber;
     property IsLocked:boolean read GetLocked;
  published
     property ConnectionPool:TkbmConnectionPool read FConnectionPool write FConnectionPool;
     property AutoSessionName:boolean read FAutoSessionName write FAutoSessionName;
     property SessionName:string read FSessionName write SetSessionName;
     property ThreadSessionName:string read FThreadSessionName;
     property OwnerID:longint read FOwnerID write FOwnerID;
  end;

  TkbmPooledSessionNameProperty = class(TEnumProperty)
    function GetValue: string; override;
    procedure GetValues(Proc: TGetStrProc); override;
    procedure SetValue(const Value: string); override;
  end;

  { TkbmPooledQueryDataLink }
  TkbmPooledQueryDataLink = class(TDataLink)
  private
    FQuery: TkbmPooledQuery;
  protected
    procedure ActiveChanged; override;
    procedure RecordChanged(Field: TField); override;
    procedure CheckBrowseMode; override;
  public
    constructor Create(AQuery: TkbmPooledQuery);
  end;

  TkbmPooledSync=(kpssAsync,kpssSync);

  TkbmPooledDataSet = class(TkbmMemTable)
  private
    FReadOnly:boolean;
    FDummy:boolean;
    FUserFields:TStringList;
    FDummyFields:TList;
    FSessionName:string;
    FSync:TkbmPooledSync;
    FRefFieldDefs:TFieldDefs;
    FCached:boolean;
    FMaxCacheRecordCount:integer;
    FIsFromCache:boolean;
    FCacheParams:TParamType;
    FCacheHits:longint;
    FCacheMisses:longint;
    FDuringDoAfterOpen:boolean;

    function GetSession:TkbmPooledSession;
  protected
    procedure Populate(CacheID:longint); virtual; abstract;
    procedure InternalOpen; override;
    procedure InternalClose; override;
    procedure DoAfterOpen; override;
    function UserFieldIndexByName(AName:string):integer;
    procedure CopyData(ds:TDataSet);
    function AcceptParamInCache(param:TParam):boolean;
    function GetUniqueID:string; virtual; abstract;
    property Cached:boolean read FCached write FCached;
    property MaxCacheRecordCount:integer read FMaxCacheRecordCount write FMaxCacheRecordCount;
    property IsFromCache:boolean read FIsFromCache;
    property CacheParams:TParamType read FCacheParams write FCacheParams;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    property Session:TkbmPooledSession read GetSession;
    property CacheHits:longint read FCacheHits;
    property CacheMisses:longint read FCacheMisses;
  published
    property ReadOnly:boolean read FReadOnly write FDummy;
    property Sync:TkbmPooledSync read FSync write FSync;
    property SessionName: string read FSessionName write FSessionName;
  end;

  TkbmPooledQuery = class(TkbmPooledDataSet)
  private
    FSQL:TStrings;
    FParams: TParams;
    FDataLink: TDataLink;
    FText: string;
    FParamCheck: Boolean;

    procedure SetParamsList(Value: TParams);
    procedure RefreshParams;
    procedure SetQuery(Value: TStrings);
    procedure CreateParams(List: TParams; const Value: PChar);
    procedure QueryChanged(Sender: TObject);
  protected
    procedure Populate(CacheID:longint); override;
    procedure InternalInitFieldDefs; override;
    function GetUniqueID:string; override;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    function ParamByName(const Value: string): TParam;
    property Text: string read FText;
    procedure ExecSQL;
    property IsFromCache;
  published
    property SQL:TStrings read FSQL write SetQuery;
    property ParamCheck: Boolean read FParamCheck write FParamCheck default True;
    property Params: TParams read FParams write SetParamsList;
    property Cached;
    property MaxCacheRecordCount;
    property CacheParams;
  end;

  TkbmPooledStoredProc = class(TkbmPooledDataSet)
  private
    FProcName:string;
    FParams: TParams;
    FBindMode:TParamBindMode;
    FOverload:word;

    procedure SetParamsList(Value: TParams);
    procedure SetProcName(AName:string);
  protected
    procedure Populate(CacheID:longint); override;
    procedure InternalInitFieldDefs; override;
    function GetUniqueID:string; override;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    function ParamByName(const Value: string): TParam;
    procedure ExecProc;
    procedure GetResults;
    property IsFromCache;
  published
    property StoredProcName: string read FProcName write SetProcName;
    property Overload: Word read FOverload write FOverload default 0;
    property Params: TParams read FParams write SetParamsList;
    property ParamBindMode: TParamBindMode read FBindMode write FBindMode default pbByName;
    property Cached;
    property CacheParams;
  end;

procedure Register;

implementation

{$IFDEF VER100} // Delphi 3
uses ParamsEditor;
{$ENDIF}

// IF YOU MOVE THIS CONST SECTION DOWN TO AFTER IMPLEMENTATION, NO PROBLEM.
// BUT IF IT IS IN THE INTERFACE SECTION THEN INTERNAL ERROR GH2300. KBM
// C++ Builder 4.0 Enterprise
const
  kbmftSupported=[ftString,ftSmallint,ftInteger,ftWord,ftBoolean,ftFloat,
                  ftCurrency,ftDate,ftTime,ftDateTime,ftAutoInc,
                  ftBCD,ftBlob,ftMemo,ftGraphic,ftFmtMemo,ftParadoxOle,
                  ftDBaseOle,ftTypedBinary,ftBytes,ftVarBytes];

{$ifdef VER100} // Delphi 3
  CRCTbl: array [0..255] of Longint = (
{$else}
  CRCTbl: array [0..255] of Cardinal = (
{$endif}
    $00000000, $77073096, $ee0e612c, $990951ba,
    $076dc419, $706af48f, $e963a535, $9e6495a3,
    $0edb8832, $79dcb8a4, $e0d5e91e, $97d2d988,
    $09b64c2b, $7eb17cbd, $e7b82d07, $90bf1d91,

    $1db71064, $6ab020f2, $f3b97148, $84be41de,
    $1adad47d, $6ddde4eb, $f4d4b551, $83d385c7,
    $136c9856, $646ba8c0, $fd62f97a, $8a65c9ec,
    $14015c4f, $63066cd9, $fa0f3d63, $8d080df5,

    $3b6e20c8, $4c69105e, $d56041e4, $a2677172,
    $3c03e4d1, $4b04d447, $d20d85fd, $a50ab56b,
    $35b5a8fa, $42b2986c, $dbbbc9d6, $acbcf940,
    $32d86ce3, $45df5c75, $dcd60dcf, $abd13d59,

    $26d930ac, $51de003a, $c8d75180, $bfd06116,
    $21b4f4b5, $56b3c423, $cfba9599, $b8bda50f,
    $2802b89e, $5f058808, $c60cd9b2, $b10be924,
    $2f6f7c87, $58684c11, $c1611dab, $b6662d3d,

    $76dc4190, $01db7106, $98d220bc, $efd5102a,
    $71b18589, $06b6b51f, $9fbfe4a5, $e8b8d433,
    $7807c9a2, $0f00f934, $9609a88e, $e10e9818,
    $7f6a0dbb, $086d3d2d, $91646c97, $e6635c01,

    $6b6b51f4, $1c6c6162, $856530d8, $f262004e,
    $6c0695ed, $1b01a57b, $8208f4c1, $f50fc457,
    $65b0d9c6, $12b7e950, $8bbeb8ea, $fcb9887c,
    $62dd1ddf, $15da2d49, $8cd37cf3, $fbd44c65,

    $4db26158, $3ab551ce, $a3bc0074, $d4bb30e2,
    $4adfa541, $3dd895d7, $a4d1c46d, $d3d6f4fb,
    $4369e96a, $346ed9fc, $ad678846, $da60b8d0,
    $44042d73, $33031de5, $aa0a4c5f, $dd0d7cc9,

    $5005713c, $270241aa, $be0b1010, $c90c2086,
    $5768b525, $206f85b3, $b966d409, $ce61e49f,
    $5edef90e, $29d9c998, $b0d09822, $c7d7a8b4,
    $59b33d17, $2eb40d81, $b7bd5c3b, $c0ba6cad,

    $edb88320, $9abfb3b6, $03b6e20c, $74b1d29a,
    $ead54739, $9dd277af, $04db2615, $73dc1683,
    $e3630b12, $94643b84, $0d6d6a3e, $7a6a5aa8,
    $e40ecf0b, $9309ff9d, $0a00ae27, $7d079eb1,

    $f00f9344, $8708a3d2, $1e01f268, $6906c2fe,
    $f762575d, $806567cb, $196c3671, $6e6b06e7,
    $fed41b76, $89d32be0, $10da7a5a, $67dd4acc,
    $f9b9df6f, $8ebeeff9, $17b7be43, $60b08ed5,

    $d6d6a3e8, $a1d1937e, $38d8c2c4, $4fdff252,
    $d1bb67f1, $a6bc5767, $3fb506dd, $48b2364b,
    $d80d2bda, $af0a1b4c, $36034af6, $41047a60,
    $df60efc3, $a867df55, $316e8eef, $4669be79,

    $cb61b38c, $bc66831a, $256fd2a0, $5268e236,
    $cc0c7795, $bb0b4703, $220216b9, $5505262f,
    $c5ba3bbe, $b2bd0b28, $2bb45a92, $5cb36a04,
    $c2d7ffa7, $b5d0cf31, $2cd99e8b, $5bdeae1d,

    $9b64c2b0, $ec63f226, $756aa39c, $026d930a,
    $9c0906a9, $eb0e363f, $72076785, $05005713,
    $95bf4a82, $e2b87a14, $7bb12bae, $0cb61b38,
    $92d28e9b, $e5d5be0d, $7cdcefb7, $0bdbdf21,

    $86d3d2d4, $f1d4e242, $68ddb3f8, $1fda836e,
    $81be16cd, $f6b9265b, $6fb077e1, $18b74777,
    $88085ae6, $ff0f6a70, $66063bca, $11010b5c,
    $8f659eff, $f862ae69, $616bffd3, $166ccf45,

    $a00ae278, $d70dd2ee, $4e048354, $3903b3c2,
    $a7672661, $d06016f7, $4969474d, $3e6e77db,
    $aed16a4a, $d9d65adc, $40df0b66, $37d83bf0,
    $a9bcae53, $debb9ec5, $47b2cf7f, $30b5ffe9,

    $bdbdf21c, $cabac28a, $53b39330, $24b4a3a6,
    $bad03605, $cdd70693, $54de5729, $23d967bf,
    $b3667a2e, $c4614ab8, $5d681b02, $2a6f2b94,
    $b40bbe37, $c30c8ea1, $5a05df1b, $2d02ef8d
  );

var
  SessionList:TkbmPooledSessionList;

{ Garbage collection routines. }

procedure TkbmPooledGarbageThread.DoGarbageCollection;
var
   f:single;
   i:integer;
   tnow:TDateTime;
   lst:TList;
begin
     with ConnectionPool do
     begin
          if not GarbageCollection then exit;

          if assigned(FOnGarbageCollection) then FOnGarbageCollection(ConnectionPool);

          tnow:=now;

          // Clean cache.
          Cache.CleanCache(MaxCacheAge,MaxCacheEntries);

          // Clean timedout connections.
          f:=FConnectionTimeout/(24*60*60);
          lst:=DBList.LockList;
          try
             // Clean timedout connections.
             for i:=lst.count-1 downto 0 do
             begin
                  with TkbmConnection(lst.Items[i]) do
                  begin
                       if (ConnectionTimeout>0) and (FLockCount<=0) and (QueueLength<=0) and (f<(tnow-LastUsed)) then
                       begin
                            free;
                            lst.Delete(i);
                       end;
                  end;
             end;
          finally
             DBList.UnlockList;
          end;
     end;
end;

procedure TkbmPooledGarbageThread.Execute;
var
   rc:integer;
begin
   repeat
         rc:=WaitForSingleObject(ConnectionPool.FStopEvent,ConnectionPool.FGarbageInterval*1000);
         if rc=WAIT_TIMEOUT then DoGarbageCollection;
   until rc<>WAIT_TIMEOUT;
end;

{ Cache handling routines. }

// Create a new cacheitem.
constructor TkbmCacheItem.Create(CacheID:longint; DataSet:TDataSet; Records:boolean);
begin
     inherited Create;
     FCacheID:=CacheID;
     FTable:=TkbmMemtable.Create(nil);
     FParams:=TParams.create;
     FCreated:=now;

     // Put data in cache.
     if Records then
        FTable.LoadFromDataSet(DataSet,true,true);

     // Put parameters in cache.
     if DataSet is TkbmPooledQuery then
        FParams.assign(TkbmPooledQuery(DataSet).Params)
     else if DataSet is TkbmPooledStoredProc then
        FParams.assign(TkbmPooledStoredProc(DataSet).Params)
     else if DataSet is TStoredProc then
        FParams.assign(TStoredProc(DataSet).Params)
     else if DataSet is TQuery then
        FParams.assign(TQuery(DataSet).Params);
end;

// Destroy cacheitem.
destructor TkbmCacheItem.Destroy;
begin
     FParams.free;
     FTable.free;
     inherited;
end;

// Create a cache list.
constructor TkbmCache.Create(AOwner:TkbmConnectionPool);
begin
     inherited Create;
     FOwner:=AOwner;
end;

// Destroy a cache list.
destructor TkbmCache.Destroy;
var
   lst:TList;
   i:integer;
begin
     lst:=LockList;
     try
        for i:=0 to lst.count-1 do
           with TkbmCacheItem(lst.Items[i]) do free;
        lst.clear;
     finally
        UnlockList;
     end;
     inherited;
end;

// Get number of entries.
function TkbmCache.GetCount:integer;
var
   lst:TList;
begin
     lst:=LockList;
     try
        Result:=lst.count;
     finally
        UnlockList;
     end;
end;

// Calc CRC32 for supplied string.
function TkbmCache.CalcCRC32(s:string):longint;
var
{$ifdef VER100} // Delphi 3
   CRCVal:Longint;
{$else}
   CRCVal:Cardinal;
{$endif}
   i,l:integer;
begin
{$ifdef VER100} // Delphi 3
     CRCVal:=high(-1);
{$else}
     CRCVal:=high(Cardinal);
{$endif}
     l:=length(s);
     for i:=1 to l do
         CRCVal:=CRCTbl[Byte(CRCVal xor Cardinal(s[i]))] xor ((CRCVal shr 8) and $00FFFFFF);
     Result:=not(CRCVal);
end;

// Look for cacheitem with specified cacheID value.
function TkbmCache.InternalLocateItem(lst:TList; CacheID:longint):integer;
var
   i:integer;
begin
     Result:=-1;
     for i:=0 to lst.Count-1 do
         if TkbmCacheItem(lst.Items[i]).FCacheID = CacheID then
         begin
              Result:=i;
              break;
         end;
end;

// Get cached result for string.
function TkbmCache.GetCached(CacheID:longint; ds:TkbmMemTable):boolean;
var
   item:TkbmCacheItem;
   lst:Tlist;
   n:integer;
begin
     Result:=false;

     // Look for item. If found copy data to dataset and return true, else return false.
     lst:=LockList;
     try
        n:=InternalLocateItem(lst,CacheID);
        if n>=0 then
        begin
             // Get the cached item.
             item:=TkbmCacheItem(lst.Items[n]);
             item.FLastAccess:=now;

             // Move entry to start of cache list.
             lst.move(n,0);

             // Get data from entry.
             ds.ReadOnly:=false;
             ds.LoadFromDataSet(item.FTable,false,false);
             if ds is TkbmPooledQuery then
                TkbmPooledQuery(ds).Params.assign(item.FParams)
             else if ds is TkbmPooledStoredProc then
                TkbmPooledStoredProc(ds).Params.assign(item.FParams);

             ds.ReadOnly:=true;

             // Get params from cache.
             Result:=true;
        end;
     finally
        UnLockList;
     end;
end;

// Add memorytable result to cache with specified searchkey.
function TkbmCache.AddToCache(CacheID:longint; ds:TDataSet; Records:boolean):TkbmCacheItem;
var
   lst:TList;
begin
     Result:=nil;

     // Check if dataset to large to cache.
     if (not FOwner.FEnableCache) or
        (Records and (FOwner.FMaxCacheRecordCount>0) and (ds.RecordCount>FOwner.FMaxCacheRecordCount)) then exit;

     // Create a cache item.
     Result:=TkbmCacheItem.Create(CacheID,ds,Records);
     Result.FLastAccess:=now;

     // Enter cache item in top of cache list.
     lst:=LockList;
     try
        lst.Insert(0,Result);
     finally
        UnLockList;
     end;
end;

// Run through cache and clean it.
procedure TkbmCache.InternalCleanCache(lst:TList; MaxAge,MaxEntries:integer);
var
   i:integer;
   item:TkbmCacheItem;
   dt:TDateTime;
   fMaxAge:single;
begin
     // Recalculate MaxAge to DateTime format.
     fMaxAge:=MaxAge/(24*60*60);

     // Clean out to old entries.
     dt:=now;
     for i:=lst.Count-1 downto 0 do
     begin
          // Get the cached item.
          item:=TkbmCacheItem(lst.Items[i]);

          // Check if to cleanout this one, as it is to old.
          if (MaxAge>0) and ((dt-item.FCreated)>fMaxAge) then
          begin
               item.Free;
               lst.Delete(i);
               exit;
          end;
     end;

     // Clean out to many entries.
     if lst.count>MaxEntries then
     begin
          for i:=lst.count-1 downto MaxEntries do
          begin
               item:=TkbmCacheItem(lst.Items[i]);
               item.free;
               lst.Delete(i);
          end;
     end;
end;

// Run through cache and clean it.
procedure TkbmCache.CleanCache(MaxAge,MaxEntries:integer);
var
   lst:TList;
begin
     // Look for item. If found, put it in top, otherwise insert a new in top.
     lst:=LockList;
     try
        InternalCleanCache(lst,MaxAge,MaxEntries);
     finally
        UnLockList;
     end;
end;

// Clear all cache entries.
procedure TkbmCache.ClearCache;
begin
     CleanCache(0,0);
end;

{ Single connection handler routines. }

constructor TkbmConnection.Create(ADatabaseTemplate:TDatabase; ConnID:integer; dbParams:TStringList);
begin
     inherited Create;
     FSession:=nil;
     FDatabase:=nil;
     FDatabase:=TDatabase.create(nil);
     FOwnerID:=-1;

     with ADatabaseTemplate do
     begin
          FDatabase.DatabaseName:=DatabaseName+inttostr(ConnID);
          FSession:=TSession.create(nil);
          FSession.SessionName:=SessionName+inttostr(ConnID);
          FDatabase.SessionName:=FSession.SessionName;
          FDatabase.LoginPrompt:=LoginPrompt;
          FDatabase.KeepConnection:=KeepConnection;
          FDatabase.HandleShared:=HandleShared;
          if AliasName<>'' then FDatabase.AliasName:=AliasName
          else FDatabase.DriverName:=DriverName;
          FDatabase.Params.assign(dbParams);
     end;
     FDatabase.Open;
     FLockCount:=0;
     FQueueLength:=0;
//     InitializeCriticalSection(QueueLock);
     FSemaphore:=CreateSemaphore(nil,1,1,nil);
end;

destructor TkbmConnection.Destroy;
begin
     if assigned(FDatabase) then
     begin
          FDatabase.close;
          FDatabase.free;
          FDatabase:=nil;
     end;

     if assigned(FSession) then
     begin
          FSession.active:=false;
          FSession.free;
          FSession:=nil;
     end;

     CloseHandle(FSemaphore);
//     DeleteCriticalSection(QueueLock);
     inherited;
end;

// Lock connection and get database.
function TkbmConnection.LockConnection(OwnerID:Longint):TDatabase;
var
// s:string;
   n:DWORD;
begin
     inc(FQueueLength);

//s:=inttostr(OwnerID)+':'+inttostr(FQueueLength)+':Trying connection lock';
//OutputDebugString(PChar(s));
     // Wait for critical section.
     inc(FLockCount);
     n:=WaitForSingleObject(FSemaphore,SEMAPHORE_TIMEOUT);
     if (n=WAIT_TIMEOUT) or (n=WAIT_FAILED) then
     begin
          Result:=nil;
//s:=inttostr(OwnerID)+':'+inttostr(FQueueLength)+':Connection lock timed failed.';
//OutputDebugString(PChar(s));
          dec(FLockCount);
          exit;
     end;

//     EnterCriticalSection(QueueLock);
//s:=inttostr(OwnerID)+':'+inttostr(FQueueLength)+':Connection lock';
//OutputDebugString(PChar(s));

     FLastUsed:=now;

     // Enter tag for who owns this now.
     FOwnerID:=OwnerID;

     Result:=FDatabase;
end;

// Unlock connection.
procedure TkbmConnection.UnlockConnection;
//var
// s:string;
begin
     if FLockCount<=0 then exit;

     dec(FQueueLength);
     FLastUsed:=now;
     FOwnerID:=-1;
     dec(FLockCount);
     ReleaseSemaphore(FSemaphore,1,nil);
//     LeaveCriticalSection(QueueLock);
//s:=inttostr(OwnerID)+':'+inttostr(FQueueLength)+':Connection unlock';
//OutputDebugString(PChar(s));
end;

function TkbmConnection.GetLocked:boolean;
begin
     Result:=FLockCount>0;
end;

// Handlers for multiple connections to a pool of databases.
constructor TkbmConnectionPool.Create(AOwner:TComponent);
begin
     inherited;
     DBList:=TThreadList.create;
     FMaxConnections:=1;
     FCache:=TkbmCache.Create(self);
     FMaxCacheEntries:=100;
     FMaxCacheAge:=0;
     FEnableCache:=false;
     FConnectionTimeout:=0;
     FUniqueConnectionID:=0;
     FGarbageCollection:=true;
     FGarbageInterval:=60;

     // Create garbage collection thread.
     FStopEvent:=CreateEvent(nil, False, False, nil);
     FGarbageThread:=TkbmPooledGarbageThread.Create(True);
     FGarbageThread.ConnectionPool:=self;
     FGarbageThread.FreeOnTerminate:=true;
     FGarbageThread.Resume;
end;

destructor TkbmConnectionPool.Destroy;
var
   i:integer;
   lst:TList;
begin
     // Close garbage collection thread.
     SetEvent(FStopEvent);
     CloseHandle(FStopEvent);

     // Clear all connections.
     lst:=DBList.LockList;
     try
        for i:=0 to lst.count-1 do
            TkbmConnection(lst.Items[i]).free;
        lst.clear;
     finally
        DBList.UnlockList;
        DBList.free;
        DBList:=nil;
     end;

     FCache.free;
     FCache:=nil;
     inherited;
end;

// Time all connections out.
procedure TkbmConnectionPool.TimeoutConnections;
var
   lst:TList;
   i:integer;
begin
     lst:=DBList.LockList;
     try
        for i:=lst.Count-1 downto 0 do
            with TkbmConnection(lst.Items[i]) do
            begin
                 FLastUsed:=0;
                 if FQueuelength<=0 then free;
                 lst.delete(i);
            end;
     finally
        DBList.UnlockList;
     end;
end;

// Kill all connections without remorse.
procedure TkbmConnectionPool.KillConnections;
var
   lst:TList;
   i:integer;
begin
     lst:=DBList.LockList;
     try
        for i:=lst.Count-1 downto 0 do
            with TkbmConnection(lst.Items[i]) do
            begin
                 free;
                 lst.delete(i);
            end;
     finally
        DBList.UnlockList;
     end;
end;

// Set garbage collection interval.
procedure TkbmConnectionPool.SetGarbageInterval(n:integer);
begin
     if (n<1) then raise ERangeError.Create('Interval must minimum be 1 sec.');
     FGarbageInterval:=n;
end;

// Return the number of connections active in the pool.
function TkbmConnectionPool.GetCurConnections:integer;
var
   lst:TList;
begin
     lst:=DBList.LockList;
     try
        Result:=lst.Count;
     finally
        DBList.UnlockList;
     end;
end;

// Return the number of requests outstanding in the pool.
function TkbmConnectionPool.GetCurRequests:integer;
var
   lst:TList;
   con:TkbmConnection;
   i:integer;
begin
     Result:=0;
     lst:=DBList.LockList;
     try
        for i:=0 to lst.Count-1 do
        begin
             con:=TkbmConnection(lst.Items[i]);
             Result:=Result+con.QueueLength;
        end;
     finally
        DBList.UnlockList;
     end;
end;

// Return a string of OwnerIDs for currently active connections.
function TkbmConnectionPool.GetCurConnectionOwnerIDs:string;
var
   lst:TList;
   con:TkbmConnection;
   i:integer;
   a:string;
begin
     Result:='';
     a:='';
     lst:=DBList.LockList;
     try
        for i:=0 to lst.Count-1 do
        begin
             con:=TkbmConnection(lst.Items[i]);
             Result:=Result+a+inttostr(i+1)+':'+inttostr(con.OwnerID);
             a:=', ';
        end;
     finally
        DBList.UnlockList;
     end;
end;

// Find most available connection in the pool.
function TkbmConnectionPool.GetBestConnection:TkbmConnection;
var
   i:integer;
   lst:TList;
   con:TkbmConnection;

   dtime:TDateTime;
   tnow:TDateTime;

   best:integer;

   tmin:TDateTime;
   qmin:integer;

//   qlen:integer;
   dbparams:TStringList;
begin
     Result:=nil;
     tnow:=now;

     // Find connection with shortest queue and shortest execution length.
     lst:=DBList.LockList;
     try
        // Look
        tmin:=Now;
        qmin:=32000;
        best:=-1;
        for i:=0 to lst.count-1 do
        begin
             con:=TkbmConnection(lst.Items[i]);

             // If empty queue, choose it.
             if con.QueueLength=0 then
             begin
                  best:=i;
                  break;
             end;

             dtime:=tnow-con.FLockedWhen;

             // If shorter queue, its the best one for now.
             if (qmin>con.QueueLength) then
             begin
                  tmin:=dtime;
                  best:=i;
                  continue;
             end

             // Else if same length as best queue, and the execution time is smallest, its the best one for now.
             else if (qmin=con.QueueLength) and (dtime<tmin) then
             begin
                  tmin:=dtime;
                  best:=i;
                  continue;
             end;
        end;

        // We didnt find any directly available connections. Is there room for adding a new?
        if (best<0) or (TkbmConnection(lst.Items[best]).QueueLength>0) and (lst.count<FMaxConnections) then
        begin
//             if best>=0 then qlen:=TkbmConnection(lst.Items[best]).QueueLength
//             else qlen:=-1;
//showmessage(format('Best=%d, queuelength=%d, Count=%d, Max=%d',[best,qlen,lst.count,FMaxConnections]));

             dbparams:=TStringlist.create;
             try
                dbparams.assign(FDBTemplate.Params);
                if assigned(FOnGetLogonParams) then
                   FOnGetLogonParams(FDBTemplate,lst.count,dbparams);
                inc(FUniqueConnectionID);
                Result:=TkbmConnection.Create(FDBTemplate,FUniqueConnectionID,dbparams);
                if Result<>nil then lst.Add(Result);
             finally
                dbparams.free;
                dbparams:=nil;
             end;
        end

        // No room for adding new connections, just use the best one we found.
        else if best>=0 then
           Result:=TkbmConnection(lst.Items[best]);

        Result.FLastUsed:=now;
        
     finally
        DBList.UnlockList;
     end;
end;

{ TkbmPooledSessionList }

constructor TkbmPooledSessionList.Create;
begin
     inherited;
     FSessions:=TThreadList.Create;
end;

destructor TkbmPooledSessionList.Destroy;
begin
     FSessions.Free;
     FSessions:=nil;
     inherited;
end;

procedure TkbmPooledSessionList.CloseAll;
var
  i:integer;
  lst:TList;
begin
  lst:=FSessions.LockList;
  try
     for i:=lst.count-1 downto 0 do
         TkbmPooledSession(lst).Free;
  finally
     FSessions.UnlockList;
  end;
end;

procedure TkbmPooledSessionList.AddSession(Session:TkbmPooledSession);
var
   lst:TList;
begin
     lst:=FSessions.LockList;
     try
        lst.Add(Session);
        inc(FUnique);
        Session.SessionNumber:=FUnique;
     finally
        FSessions.UnlockList;
     end;
end;

procedure TkbmPooledSessionList.DelSession(Session:TkbmPooledSession);
var
   lst:TList;
   i:integer;
begin
     if self=nil then exit;

     lst:=FSessions.LockList;
     try
        i:=lst.IndexOf(Session);
        if i>=0 then lst.Delete(i);
     finally
        FSessions.UnlockList;
     end;
end;

// Find session in list by name.
function TkbmPooledSessionList.SessionByName(const SessionName:string):TkbmPooledSession;
var
   i:integer;
   lst:TList;
begin
     Result:=nil;
     lst:=FSessions.LockList;
     try
        for i:=0 to lst.count-1 do
            if TKbmPooledSession(lst.Items[i]).ThreadSessionName=SessionName then
            begin
                Result:=TKbmPooledSession(lst.Items[i]);
                break;
            end;
     finally
        FSessions.UnlockList;
     end;
end;

{ TkbmPooledSession }

constructor TkbmPooledSession.Create(AOwner:TComponent);
begin
     inherited;

     // Allocate space.
     FSemaphore:=CreateSemaphore(nil,1,1,nil);
     FLockCount:=0;
     FOwnerID:=-1;

     // Add to sessionlist.
     SessionList.AddSession(self);

     SessionName:=Name;
end;

destructor TkbmPooledSession.Destroy;
begin
     // Remove from sessionlist.
     SessionList.DelSession(self);

     // Free up space used.
     CloseHandle(FSemaphore);

     inherited;
end;

procedure TkbmPooledSession.SetName(const NewName: TComponentName);
begin
  inherited SetName(NewName);
  UpdateThreadSessionName;
end;

procedure TkbmPooledSession.Notification(AComponent: TComponent; Operation: TOperation);
begin
     inherited;

     // Check if to change session connection on inserted pooled dataset.
     if AutoSessionName and (Operation = opInsert) then
        if (AComponent is TkbmPooledDataSet) and
          ((AComponent as TkbmPooledDataSet).SessionName=SessionName) then
             TkbmPooledDataSet(AComponent).SessionName := Self.ThreadSessionName;
end;

procedure TkbmPooledSession.SetSessionName(SessionName:string);
begin
     FSessionName:=SessionName;
     UpdateThreadSessionName;
end;

procedure TkbmPooledSession.UpdateThreadSessionName;
begin
     if AutoSessionName and (not (csDesigning in ComponentState)) then
         FThreadSessionName:=FSessionName+'_'+inttostr(SessionNumber)
     else
         FThreadSessionName:=FSessionName;
end;

procedure TkbmPooledSession.Loaded;
begin
     inherited;
     SetSessionNames;
end;

procedure TkbmPooledSession.SetSessionNames;
var
   i:integer;
   c:TComponent;
begin
    if Owner <> nil then
       for i:=0 to Owner.ComponentCount-1 do
       begin
            c:=Owner.Components[i];
            if (c is TkbmPooledDataSet) then
            begin
//    ShowMessage('Name='+c.Name+' SessionName='+TkbmPooledDataSet(c).SessionName+' Session='+Self.SessionName);
               if (AnsiCompareText(TkbmPooledDataSet(c).SessionName, Self.SessionName)=0) then
                  TkbmPooledDataSet(c).SessionName:=Self.ThreadSessionName;
            end;
        end;
end;

function TkbmPooledSession.LockSession(Sync:boolean):TkbmConnection;
var
//   s:string;
   n:DWORD;
begin
     assert(Assigned(ConnectionPool),'Connectionpool not specified in '+Name);

     // If asyncrone, just give dataset a database connection.
     if not Sync then
     begin
          Result:=ConnectionPool.GetBestConnection;
          exit;
     end;

//s:=inttostr(OwnerID)+':Trying session lock';
//OutputDebugString(PChar(s));
     // Otherwise its a syncrone call. Wait until session is not busy and lock it.
     inc(FLockCount);
     n:=WaitForSingleObject(FSemaphore,SEMAPHORE_TIMEOUT);
     if (n=WAIT_TIMEOUT) or (n=WAIT_FAILED) then
     begin
          Result:=nil;
//s:=inttostr(OwnerID)+':Session lock failed';
//OutputDebugString(PChar(s));
          dec(FLockCount);
          exit;
     end;
//s:=inttostr(OwnerID)+':Session locked';
//OutputDebugString(PChar(s));

     // Get connection.
     FCon:=ConnectionPool.GetBestConnection;
     Result:=FCon;
end;

procedure TkbmPooledSession.UnlockSession;
//var
//   s:string;
begin
     if FLockCount<=0 then exit;

     // Unlock session.
     dec(FLockCount);
     ReleaseSemaphore(FSemaphore,1,nil);
//s:=inttostr(OwnerID)+':Session unlocked';
//OutputDebugString(PChar(s));
end;

function TkbmPooledSession.GetLocked:boolean;
begin
     Result:=FLockCount>0;
end;

function TkbmPooledSessionNameProperty.GetValue: string;
begin
     Result:=GetStrValue;
end;

procedure TkbmPooledSessionNameProperty.GetValues(Proc: TGetStrProc);
var
   i:integer;
   lst:TList;
begin
     lst:=SessionList.FSessions.LockList;
     try
        for i:=0 to lst.count-1 do
            Proc(TkbmPooledSession(lst.Items[i]).SessionName);
     finally
        SessionList.FSessions.UnlockList;
     end;
end;

procedure TkbmPooledSessionNameProperty.SetValue(const Value: string);
begin
     SetStrValue(Value);
end;

{ TkbmPooledQueryDataLink }

constructor TkbmPooledQueryDataLink.Create(AQuery: TkbmPooledQuery);
begin
  inherited Create;
  FQuery := AQuery;
end;

procedure TkbmPooledQueryDataLink.ActiveChanged;
begin
  if FQuery.Active then FQuery.RefreshParams;
end;

procedure TkbmPooledQueryDataLink.RecordChanged(Field: TField);
begin
  if (Field = nil) and FQuery.Active then FQuery.RefreshParams;
end;

procedure TkbmPooledQueryDataLink.CheckBrowseMode;
begin
  if FQuery.Active then FQuery.CheckBrowseMode;
end;

{ Pooled dataset. }

constructor TkbmPooledDataSet.Create(AOwner:TComponent);
begin
     inherited;
     FReadOnly:=true;
     FDuringDoAfterOpen:=false;
     FUserFields:=TStringList.Create;
     FDummyFields:=TList.Create;
     FRefFieldDefs:=TFieldDefs.Create(self);
     FCached:=false;
end;

destructor TkbmPooledDataSet.Destroy;
begin
     Close;
     FRefFieldDefs.free;
     FUserFields.free;
     FDummyFields.free;
     inherited;
end;

function TkbmPooledDataSet.GetSession:TkbmPooledSession;
begin
     Result:=SessionList.SessionByName(SessionName);
end;

procedure TkbmPooledDataSet.DoAfterOpen;
var
   CacheID:longint;
begin
     // Check if during doafteropen, dont reenter.
     if FDuringDoAfterOpen then exit;

     // Wait for this object to be available. Lock it for our use.
     assert(Session<>nil,'Session must be specified.');

     // Check if cacheinfo available.
     FDuringDoAfterOpen:=true;
     try
         if FCached then
         begin
              // Look for a cache entry. If none, connect to DB and add to cache.
              with Session do
              begin
                   CacheID:=ConnectionPool.Cache.CalcCRC32(self.GetUniqueID);
                   if ConnectionPool.Cache.GetCached(CacheID,self) then
                   begin
                        inc(ConnectionPool.FCacheHits);
                        inc(FCacheHits);
                        FIsFromCache:=true;
                        exit;
                   end;
              end;
         end
         else
             CacheID:=0;

         // Get data from DB and cache if set.
         Populate(CacheID);
         if FCached then
         begin
              inc(Session.ConnectionPool.FCacheMisses);
              inc(FCacheMisses);
         end;
         FIsFromCache:=false;

         inherited;
     finally
         FDuringDoAfterOpen:=false;
     end;
end;

function TkbmPooledDataSet.UserFieldIndexByName(AName:string):integer;
var
   i:integer;
   f:TField;
begin
     Result:=-1;
     for i:=0 to FUserFields.Count-1 do
     begin
          f:=TField(FUserFields.Objects[i]);
          if f.FieldName=AName then
          begin
               Result:=i;
               break;
          end;
     end;
end;

procedure TkbmPooledDataSet.InternalClose;
var
   i:integer;
begin
     BindFields(False);
     if not DefaultFields then
     begin
          // Remove dummyfields from official field list.
          for i:=FDummyFields.count-1 downto 0 do
          begin
               with TField(FDummyFields.Items[i]) do
               begin
                    DataSet:=nil;
                    free;
               end;
          end;
          FDummyFields.Clear;

          // Remove user field list. (BUT DONT FREE FIELDS!!!!)
          FUserFields.Clear;
     end;

     inherited;
end;

procedure TkbmPooledDataSet.InternalOpen;
begin
     if FSessionName='' then raise Exception.Create('SessionName must be given.');
     inherited;
end;

procedure TkbmPooledDataSet.CopyData(ds:TDataSet);
var
   i:integer;
   fc:integer;
   f:TField;
   fi:array [0..255] of integer;
begin
     try
         fc:=FieldCount-1;

         // Build name index relations between memory table and actual dataset.
         for i:=0 to fc do
         begin
              if (Fields[i].FieldKind<>fkData) or (not (Fields[i].DataType in kbmftSupported)) then
              begin
                   fi[i]:=-1;
                   continue;
              end;

              f:=ds.FindField(Fields[i].FieldName);
              if f=nil then
              begin
                   fi[i]:=-1;
                   continue;
              end;

              fi[i]:=f.Index;
         end;

         // Copy data.
         inherited ReadOnly:=false;
         EmptyTable;
         ds.first;
         while not ds.EOF do
         begin
              append;
              for i:=0 to fc do
              begin
                   if fi[i]>=0 then
                   begin
                       f:=ds.Fields[fi[i]];
                       Fields[i].Value:=f.Value;

(*
                       case f.DataType of
                            ftSmallInt,
                            ftInteger,
                            ftWord,
                            ftAutoInc,
                            ftBoolean:     Fields[i].asinteger:=f.AsInteger;
                            ftString:      Fields[i].asstring:=f.AsString;
                            ftDate,
                            ftTime,
                            ftDateTime:    Fields[i].AsDateTime:=f.AsDateTime;
                            ftFloat,
                            ftCurrency:    Fields[i].AsFloat:=f.AsFloat;
                       end;
*)
                   end
                   else Fields[i].Clear;
              end;
              post;
              ds.next;
         end;
         first;
     finally
         inherited ReadOnly:=true;
     end;
end;

// Return true if param should be accounted for in cache.
function TkbmPooledDataset.AcceptParamInCache(Param:TParam):boolean;
var
   pt:TParamType;
begin
     if FCacheParams=ptUnknown then
        Result:=true
     else
     begin
          pt:=Param.ParamType;
          if (pt=CacheParams) or
             ((CacheParams=ptInputOutput) and ((pt=ptInput) or (pt=ptOutput))) then
             Result:=true
          else
             Result:=false;
     end;
end;

{ Pooled query. }

constructor TkbmPooledQuery.Create(AOwner:TComponent);
begin
     inherited;
     FSQL := TStringList.Create;
     TStringList(SQL).OnChange := QueryChanged;
{$ifndef VER100} // Not Delphi 3
     FParams := TParams.Create(self);
{$else}
     FParams := TParams.Create;
{$endif}
     FDataLink := TkbmPooledQueryDataLink.Create(Self);
end;

destructor TkbmPooledQuery.Destroy;
begin
     SQL.Free;
     FParams.Free;
     FDataLink.Free;
     inherited;
end;

procedure TkbmPooledQuery.InternalInitFieldDefs;
var
   Query:TQuery;
   DB:TDatabase;
   Con:TkbmConnection;
begin
     assert(Session<>nil,'Session must be specified.');

     // Get a pooled connection and hold on to it until finished.
     Con:=Session.LockSession(FSync=kpssSync);
     if Con=nil then raise EDatabaseError.Create('PooledQuery : InternalInitFieldDefs : Timeout/error waiting for session: '+Session.SessionName);
     try
        DB:=Con.LockConnection(Session.OwnerID);
        try
           // Create a query to handle the actual query.
           Query:=TQuery.create(self);
           try
              Query.SQL.Assign(SQL);
              Query.Params.Assign(Params);
              Query.SessionName:=DB.SessionName;
              Query.DatabaseName:=DB.DatabaseName;
              Query.RequestLive:=false;
              Query.CachedUpdates:=false;
              Query.FieldDefs.Update;

              // Copy the layout to the memorytable.
              FieldDefs.Clear;
              FieldDefs.Assign(Query.FieldDefs);
           finally
              Query.free;
           end;

        finally
           Con.UnlockConnection;
        end;
     finally
        Session.UnlockSession;
     end;
end;

function TkbmPooledQuery.GetUniqueID:string;
var
   i:integer;
begin
     Result:=trim(SQL.Text)+',';

     // Build dependability on params.
     for i:=0 to Params.Count-1 do
         if AcceptParamInCache(Params[i]) then Result:=Result+Params[i].AsString+',';
end;

procedure TkbmPooledQuery.Populate(CacheID:longint);
var
   Query:TQuery;
   DB:TDatabase;
   Con:TkbmConnection;
   DoCache:boolean;
begin
     assert(Session<>nil,'Session must be specified.');

     // Get a pooled connection and hold on to it until finished.
     DisableControls;
     DoCache:=false;
     Con:=Session.LockSession(FSync=kpssSync);
     if Con=nil then raise EDatabaseError.Create('PooledQuery : Populate : Timeout/error waiting for session: '+Session.SessionName);
     try
        DB:=Con.LockConnection(Session.OwnerID);
        try

           // Create a query to handle the actual query.
           Query:=TQuery.create(self);
           try
              Query.SQL.Assign(SQL);
              Query.Params.Assign(Params);
              Query.SessionName:=DB.SessionName;
              Query.DatabaseName:=DB.DatabaseName;
              Query.RequestLive:=false;
              Query.CachedUpdates:=false;
              Query.Open;

              // Copy the contents to the memorytable.
              Params.Assign(Query.Params);
              CopyData(Query);
              DoCache:=FCached and ((FMaxCacheRecordCount<=0) or (Query.RecordCount<FMaxCacheRecordCount));
           finally
              Query.free;
           end;

        finally
           Con.UnlockConnection;
        end;

     finally
        Session.UnlockSession;

        // Copy contents to cache.
        if DoCache then
           Session.ConnectionPool.Cache.AddToCache(CacheID,self,true);

        EnableControls;
     end;
end;

procedure TkbmPooledQuery.ExecSQL;
var
   query:TQuery;
   db:TDatabase;
   con:TkbmConnection;
   CacheID:longint;
   ses:TkbmPooledSession;
begin
     Close;
     assert(Session<>nil,'Session must be specified.');
     CacheID:=0;                // To satisfy compiler and avoid warnings although totally unneeded.
     ses:=Session;

     // Check if cacheinfo available.
     if FCached then
     begin
          // Look for a cache entry. If none, connect to DB and add to cache.
          CacheID:=ses.ConnectionPool.Cache.CalcCRC32(self.GetUniqueID);
          if ses.ConnectionPool.Cache.GetCached(CacheID,self) then
          begin
               FIsFromCache:=true;
               inc(ses.ConnectionPool.FCacheHits);
               inc(FCacheHits);
               exit;
          end;
     end;

     // Get a pooled connection and hold on to it until finished.
     con:=ses.LockSession(FSync=kpssSync);
     if Con=nil then raise EDatabaseError.Create('PooledQuery : ExecSQL : Timeout/error waiting for session: '+Session.SessionName);
     try
        db:=con.LockConnection(Session.OwnerID);
        try
           // Create a query to handle the actual query.
           query:=TQuery.create(self);
           try
             query.SQL.Assign(SQL);
             query.Params.Assign(Params);
             with query do
             begin
                  SessionName:=db.SessionName;
                  DatabaseName:=db.DatabaseName;
                  RequestLive:=false;
                  CachedUpdates:=false;
                  ExecSQL;
             end;
             Params.Assign(query.Params);
           finally
             query.free;
           end;
        finally
           con.UnlockConnection;
        end;
     finally
        Session.UnlockSession;

        // Add result to cache.
        if FCached then
        begin
             ses.ConnectionPool.Cache.AddToCache(CacheID,self,false);
             inc(ses.ConnectionPool.FCacheMisses);
             inc(FCacheMisses);
        end;
     end;
end;

// When sql statement changed, rebuild parameter list etc.
procedure TkbmPooledQuery.QueryChanged(Sender: TObject);
var
  List: TParams;
begin
  FText := SQL.Text;
  if not (csLoading in ComponentState) then
  begin
    Close;
    if ParamCheck or (csDesigning in ComponentState) then
    begin
{$ifndef VER100} // Not Delphi 3
      List := TParams.Create(self);
{$else}
      List := TParams.Create;
{$endif}
      try
        CreateParams(List, PChar(Text));
        List.AssignValues(FParams);
        FParams.Free;
        FParams := List;
      except
        List.Free;
      end;
    end;
    DataEvent(dePropertyChange, 0);
  end else
    CreateParams(nil, PChar(Text));
end;

// Set SQL statement.
procedure TkbmPooledQuery.SetQuery(Value: TStrings);
begin
  if SQL.Text <> Value.Text then
  begin
    Close;
    SQL.BeginUpdate;
    try
      SQL.Assign(Value);
    finally
      SQL.EndUpdate;
    end;
  end;
end;

procedure TkbmPooledQuery.SetParamsList(Value: TParams);
begin
  FParams.AssignValues(Value);
end;

procedure TkbmPooledQuery.RefreshParams;
var
  DataSet: TDataSet;
begin
  DisableControls;
  try
    if FDataLink.DataSource <> nil then
    begin
      DataSet := FDataLink.DataSource.DataSet;
      if DataSet <> nil then
        if DataSet.Active and (DataSet.State <> dsSetKey) then
        begin
          Close;
          Open;
        end;
    end;
  finally
    EnableControls;
  end;
end;

function TkbmPooledQuery.ParamByName(const Value: string): TParam;
begin
  Result := FParams.ParamByName(Value);
end;

procedure TkbmPooledQuery.CreateParams(List: TParams; const Value: PChar);
var
  CurPos, StartPos: PChar;
  CurChar: Char;
  Literal: Boolean;
  EmbeddedLiteral: Boolean;
  Name: string;

  function NameDelimiter: Boolean;
  begin
    Result := CurChar in [' ', ',', ';', ')', #13, #10];
  end;

  function IsLiteral: Boolean;
  begin
    Result := CurChar in ['''', '"'];
  end;

  function StripLiterals(Buffer: PChar): string;
  var
    Len: Word;
    TempBuf: PChar;

    procedure StripChar(Value: Char);
    begin
      if TempBuf^ = Value then
        StrMove(TempBuf, TempBuf + 1, Len - 1);
      if TempBuf[StrLen(TempBuf) - 1] = Value then
        TempBuf[StrLen(TempBuf) - 1] := #0;
    end;

  begin
    Len := StrLen(Buffer) + 1;
    TempBuf := AllocMem(Len);
    Result := '';
    try
      StrCopy(TempBuf, Buffer);
      StripChar('''');
      StripChar('"');
      Result := StrPas(TempBuf);
    finally
      FreeMem(TempBuf, Len);
    end;
  end;

begin
  CurPos := Value;
  Literal := False;
  EmbeddedLiteral := False;
  repeat
    CurChar := CurPos^;
    if (CurChar = ':') and not Literal and ((CurPos + 1)^ <> ':') then
    begin
      StartPos := CurPos;
      while (CurChar <> #0) and (Literal or not NameDelimiter) do
      begin
        Inc(CurPos);
        CurChar := CurPos^;
        if IsLiteral then
        begin
          Literal := Literal xor True;
          if CurPos = StartPos + 1 then EmbeddedLiteral := True;
        end;
      end;
      CurPos^ := #0;
      if EmbeddedLiteral then
      begin
        Name := StripLiterals(StartPos + 1);
        EmbeddedLiteral := False;
      end
      else Name := StrPas(StartPos + 1);
      if Assigned(List) then
        List.CreateParam(ftUnknown, Name, ptUnknown);
      CurPos^ := CurChar;
      StartPos^ := '?';
      Inc(StartPos);
      StrMove(StartPos, CurPos, StrLen(CurPos) + 1);
      CurPos := StartPos;
    end
    else if (CurChar = ':') and not Literal and ((CurPos + 1)^ = ':') then
      StrMove(CurPos, CurPos + 1, StrLen(CurPos) + 1)
    else if IsLiteral then Literal := Literal xor True;
    Inc(CurPos);
  until CurChar = #0;
end;

{ Pooled stored procedure handler routines. }
constructor TkbmPooledStoredProc.Create(AOwner:TComponent);
begin
     inherited;
{$ifndef VER100} // Not Delphi 3
     FParams := TParams.Create(self);
{$else}
     FParams := TParams.Create;
{$endif}
end;

destructor TkbmPooledStoredProc.Destroy;
begin
//{$ifdef VER100} // Delphi 3
     FParams.Free;
//{$endif}
     inherited;
end;

procedure TkbmPooledStoredProc.InternalInitFieldDefs;
var
   Proc:TStoredProc;
   DB:TDatabase;
   Con:TkbmConnection;
begin
     assert(Session<>nil,'Session must be specified.');

     // Get a pooled connection and hold on to it until finished.
     Con:=Session.LockSession(FSync=kpssSync);
     if Con=nil then raise EDatabaseError.Create('PooledStoredProc : InternalInitFieldDefs : Timeout/error waiting for session: '+Session.SessionName);
     try
        DB:=Con.LockConnection(Session.OwnerID);
        try
           // Create a query to handle the actual query.
           Proc:=TStoredProc.create(self);
           try
              Proc.SessionName:=DB.SessionName;
              Proc.DatabaseName:=DB.DatabaseName;
              Proc.StoredProcName:=StoredProcName;
              Proc.Params.Assign(Params);
              Proc.Overload:=FOverload;
              Proc.Prepare;

              // Copy the layout to the memorytable.
              FieldDefs.Clear;
              FieldDefs.Assign(Proc.FieldDefs);
              Proc.CopyParams(FParams);
           finally
              Proc.UnPrepare;
              Proc.free;
           end;

        finally
           Con.UnlockConnection;
        end;
     finally
        Session.UnlockSession;
     end;
end;

function TkbmPooledStoredProc.GetUniqueID:string;
var
   i:integer;
begin
     Result:=StoredProcName+','+inttostr(FOverLoad)+',';

     // Build dependability on params.
     for i:=0 to Params.Count-1 do
         if AcceptParamInCache(Params[i]) then Result:=Result+Params[i].AsString+',';
end;

procedure TkbmPooledStoredProc.Populate(CacheID:longint);
var
   Proc:TStoredProc;
   DB:TDatabase;
   Con:TkbmConnection;
   DoCache:boolean;
begin
     assert(Session<>nil,'Session must be specified.');

     // Get a pooled connection and hold on to it until finished.
     DisableControls;
     DoCache:=false;
     Con:=Session.LockSession(FSync=kpssSync);
     if Con=nil then raise EDatabaseError.Create('PooledQuery : Populate : Timeout/error waiting for session: '+Session.SessionName);
     try
        DB:=Con.LockConnection(Session.OwnerID);
        try
           Proc:=TStoredProc.create(self);
           try
              with Proc do
              begin
                   SessionName:=db.SessionName;
                   DatabaseName:=db.DatabaseName;
              end;
              Proc.StoredProcName:=StoredProcName;
              Proc.Params.Assign(Params);
              Proc.Overload:=FOverload;
              Proc.ExecProc;
              Proc.GetResults;

              // Copy the contents to the memory table.
              FParams.Assign(Proc.Params);

              CopyData(Proc);
              DoCache:=FCached and ((FMaxCacheRecordCount<=0) or (Proc.RecordCount<FMaxCacheRecordCount));
           finally
              Proc.Close;
              Proc.Free;
           end;

        finally
           Con.UnlockConnection;
        end;

     finally
        Session.UnlockSession;

        // Copy contents to cache.
        if DoCache then
           Session.ConnectionPool.Cache.AddToCache(CacheID,self,true);

        EnableControls;
     end;
end;

procedure TkbmPooledStoredProc.ExecProc;
var
   Proc:TStoredProc;
   DB:TDatabase;
   Con:TkbmConnection;
   CacheID:longint;
   ses:TkbmPooledSession;
begin
     Close;
     assert(Session<>nil,'Session must be specified.');
     CacheID:=0;                // To satisfy compiler and avoid warnings although totally unneeded.
     ses:=Session;

     // Check if cacheinfo available.
     if FCached then
     begin
          // Look for a cache entry. If none, connect to DB and add to cache.
          CacheID:=ses.ConnectionPool.Cache.CalcCRC32(self.GetUniqueID);
          if ses.ConnectionPool.Cache.GetCached(CacheID,self) then
          begin
               inc(ses.ConnectionPool.FCacheHits);
               inc(FCacheHits);
               FIsFromCache:=true;
               exit;
          end;
     end;

     // Get connection to database and execute stored procedure.
     Con:=Session.LockSession(FSync=kpssSync);
     if Con=nil then raise EDatabaseError.Create('PooledStoredProc : ExecProc : Timeout/error waiting for session: '+Session.SessionName);
     try
        DB:=con.LockConnection(Session.OwnerID);
        try
           // Create a stored procedure to handle the actual call.
           Proc:=TStoredProc.create(self);
           try
              with Proc do
              begin
                   SessionName:=db.SessionName;
                   DatabaseName:=db.DatabaseName;
              end;
              Proc.StoredProcName:=StoredProcName;
              Proc.Params.Assign(Params);
              Proc.Overload:=FOverload;
              Proc.ExecProc;
              Proc.GetResults;

              // Copy the contents to the memory table.
              FParams.Assign(Proc.Params);
           finally
              Proc.close;
              Proc.free;
           end;

        finally
           Con.UnLockConnection;
        end;
     finally
        Session.UnlockSession;

        // Add result to cache.
        if FCached then
        begin
             Session.ConnectionPool.Cache.AddToCache(CacheID,self,false);
             inc(ses.ConnectionPool.FCacheMisses);
             inc(FCacheMisses);
        end;
     end;
end;

procedure TkbmPooledStoredProc.GetResults;
begin
end;

procedure TkbmPooledStoredProc.SetProcName(AName:string);
begin
     FProcName:=AName;
     if AName='' then
     begin
          FParams.clear;
          FieldDefs.clear;
     end
     else if not (csLoading in ComponentState) then
        InternalInitFieldDefs;
end;

procedure TkbmPooledStoredProc.SetParamsList(Value: TParams);
begin
    FParams.Assign(Value);
end;

function TkbmPooledStoredProc.ParamByName(const Value: string): TParam;
begin
  Result := FParams.ParamByName(Value);
end;

//*****************************************************
// Parameter editor (Delphi 3 support).
//*****************************************************

{$IFDEF VER100} // Delphi 3
type
    { TParamsProperty }
    TkbmParamsProperty = class(TPropertyEditor)
    public
       function GetValue: string; override;
       function GetAttributes: TPropertyAttributes; override;
    end;

    { TQueryParamsProperty }
    TkbmPooledDatasetParamsProperty = class(TkbmParamsProperty)
    public
       procedure Edit; override;
    end;

procedure TkbmPooledDatasetParamsProperty.Edit;
var
   List1,List2: TParams;
   ds: TkbmPooledDataset;
begin
     ds:=GetComponent(0) as TkbmPooledDataset;
     List1 := TParams.Create;
     List2 := TParams.Create;
     try
        if (ds is TkbmPooledQuery) then
           List1.Assign(TkbmPooledQuery(ds).Params)
        else if (ds is TkbmPooledStoredProc) then
           List1.Assign(TkbmPooledStoredProc(ds).Params)
        else
            DatabaseError('Params-Edit: Unknown TkbmPooledDataset descendant.');
        List2.Assign(List1);
        if EditParams(List1) and not List1.IsEqual(List2) then
        begin
             if (ds is TkbmPooledQuery) then
                TkbmPooledQuery(ds).Params:=List1
             else if (ds is TkbmPooledStoredProc) then
                TkbmPooledStoredProc(ds).Params:=List1;
             Modified;
        end;

     finally
        List1.Free;
        List2.Free;
     end;
end;

function TkbmParamsProperty.GetValue: string;
begin
  Result := Format('(%s)', [TParams.ClassName]);
end;

function TkbmParamsProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paMultiSelect, paDialog];
end;
{$endif}

procedure Register;
begin
  RegisterComponents('Data Access', [TkbmConnectionPool,TkbmPooledSession,TkbmPooledQuery,TkbmPooledStoredProc]);
  RegisterPropertyEditor(TypeInfo(string), TkbmPooledDataSet, 'SessionName', TkbmPooledSessionNameProperty);

{$IFDEF VER100} // Delphi 3
  RegisterPropertyEditor(TypeInfo(TParams), TkbmPooledDataSet, 'Params', TkbmPooledDatasetParamsProperty);
{$endif}
end;

initialization
   SessionList:=TkbmPooledSessionList.create;

finalization
   SessionList.free;
   SessionList:=nil;

end.





